#include "Framework.h"
#include "MapClass.h"
#include <vector>
#include <cmath>

using namespace std;

MapObj::MapObj()
{
  Start=End=-1;
}

// Populate map hexes
void MapObj::Init(int Rows, int Cols)
{
  int InitLoop;
  HexStruct NewHex;

  // Build a hex with neutral attributes
  NewHex.Type=TYPE_NORMAL;
  NewHex.List=0;
  NewHex.Previous=-1;
  NewHex.F=-1;
  NewHex.G=0.0f;
  NewHex.H=0.0f;

  HexCount=Rows*Cols;
  RowPitch=Rows;

  // Populate hex vector
  for(InitLoop=0;InitLoop!=HexCount;++InitLoop)
    Hex.push_back(NewHex);
}


// Normal Render
void MapObj::Render()
{
  int RowDex,ColDex;

  glPushMatrix(); // Save starting matrix

  // Set hex color based on Type and List properties
  for(HexItr=Hex.begin(),RowDex=1,ColDex=0;HexItr!=Hex.end();++HexItr,++RowDex)
  {
    switch(HexItr->Type)
    {
    case TYPE_BLOCKED:
      glColor3f(1.0f,0.0f,0.0f);
      break;

    case TYPE_NORMAL:
      glColor3f(0.5f,0.75f,0.0f);
      break;

    case TYPE_START:
      glColor3f(0.0f,1.0f,0.0f);
      break;

    case TYPE_FINISH:
      glColor3f(0.0f,0.0f,1.0f);
      break;
    }

    switch(HexItr->List)
    {
    case LIST_OPEN:
      glColor3f(0.6f,0.6f,0.6f);
      break;

    case LIST_CLOSED:
      glColor3f(0.25f,0.25f,0.25f);
      break;

    case LIST_PATH:
      glColor3f(1.0f,1.0f,0.0f);
      break;
    }

    switch(HexItr->Type)
    {
    case TYPE_START:
      glColor3f(0.0f,1.0f,0.0f);
      break;

    case TYPE_FINISH:
      glColor3f(0.0f,0.0f,1.0f);
      break;
    }

    // Draw the hex
    glBegin(GL_POLYGON);
    glVertex3f(-0.5f,0.0f, 0.72f);
    glVertex3f( 0.5f,0.0f, 0.72f);
    glVertex3f( 1.0f,0.0f, 0.0f);
    glVertex3f( 0.5f,0.0f,-0.72f);
    glVertex3f(-0.5f,0.0f,-0.72f);
    glVertex3f(-1.0f,0.0f, 0.0f);
    glEnd();

    // Move the the next hex position
    glTranslatef(0.0f,0.0f,HEX_Z_OFFSET);

    // Handle row and column adjustments
    if(RowDex==RowPitch)
    {
      RowDex=0;
      ++ColDex;
      glPopMatrix(); // Reset to starting matrix
      glPushMatrix();
      glTranslatef(HEX_X_OFFSET*(float)ColDex,0.0f,0.0f); // Move across for next column
      if(ColDex%2)
        glTranslatef(0.0f,0.0f,HEX_Z_OFFSET/-2.0f);       // Add an offset for odd numbered rows
    }

  }

  glPopMatrix(); // Restore starting matrix
}

// Render with cheap selection
void MapObj::Render(bool Select)
{
 int RowDex,ColDex;
 char RedVal,GreenVal;

  if(!Select) // If selection is not required, do a normal render
  {
    Render();
    return;
  }

  glPushMatrix();  // Save Starting matrix

  // Draw each hex with an incrementing red value.
  // Green values are incremented when the red value gets to 128
   for(HexItr=Hex.begin(),RedVal=1,RowDex=1,ColDex=0,GreenVal=0;
       HexItr!=Hex.end();++HexItr,++RowDex,++RedVal)
    {
       if(RedVal==127)
        {
         RedVal=1;
         ++GreenVal;
        }
        
      glColor3b(RedVal,GreenVal,64);

      glBegin(GL_POLYGON);
      glVertex3f(-0.5f,0.0f, 0.72f);
      glVertex3f( 0.5f,0.0f, 0.72f);
      glVertex3f( 1.0f,0.0f, 0.0f);
      glVertex3f( 0.5f,0.0f,-0.72f);
      glVertex3f(-0.5f,0.0f,-0.72f);
      glVertex3f(-1.0f,0.0f, 0.0f);
      glEnd();

      // Move to the next hex position
      glTranslatef(0.0f,0.0f,HEX_Z_OFFSET);

      // Handle row and column adjusments
      if(RowDex==RowPitch)
      {
        RowDex=0;
        ++ColDex;
        glPopMatrix();  // Reset to start matrix
        glPushMatrix();
        glTranslatef(HEX_X_OFFSET*(float)ColDex,0.0f,0.0f); // Move across for next column
        if(ColDex%2)
          glTranslatef(0.0f,0.0f,HEX_Z_OFFSET/-2.0f);       // Add an offset for odd columns
      }
    }
    
  glPopMatrix(); // Restore start matrix
}

// Set hex ensures that there is only one start ad end point.
// It also clears the list attributes of the map when any hex is modified.
void MapObj::SetHexType(int Num, int Type)
{
  // Clear list attributes
  for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
  {
    HexItr->List=LIST_NONE;
  }

  if(Num<=HexCount) // Ensure we're not out of range
  {
    // Clear existing start hexes
    if(Type==TYPE_START)
    {
      Start=Num;
      for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
      {
        if(HexItr->Type==TYPE_START)
          HexItr->Type=TYPE_NORMAL;
      }
    }

    // Clear existing end hexes
    if(Type==TYPE_FINISH)
    {
      End=Num;
      for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
      {
        if(HexItr->Type==TYPE_FINISH)
          HexItr->Type=TYPE_NORMAL;
      }
    }

    // Set hex type
    Hex[Num].Type=Type;
  }
}


// Reset hex list and type attributes
void MapObj::Reset()
{
  Start=End=-1;

  for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
  {
    HexItr->Type=TYPE_NORMAL;
    HexItr->List=LIST_NONE;
  }
}


// Get Row and Column position from a hex index
HexPosition MapObj::GetPos(int HexNum)
{
  HexPosition Result;

  Result.Col=-1;
  Result.Row=-1;

  if(RowPitch) // Prevent divide by zero
  {
    Result.Col=(int)HexNum/RowPitch;
    Result.Row=HexNum-(Result.Col*RowPitch);
  }

  return Result;
}


// Get the coordinates of a hex,
// this is required for the calculation of the H value
HexCoordinate MapObj::GetCoord(int HexNum)
{
  HexCoordinate HexCoord;
  HexPosition HexPos;

  // Find Row and Column
  HexPos=GetPos(HexNum);

  // Calculate coords
  HexCoord.x=HEX_X_OFFSET * HexPos.Col;
  HexCoord.z=HEX_Z_OFFSET * HexPos.Row;

  // Add offset for odd columns
  if(HexPos.Col%2)
    HexCoord.z-=HEX_Z_OFFSET/2.0f;

  return HexCoord;
}


// A bit of trig to calculate the distance between hexes
// The supplied hex indices are converted to coordinates
// The distance between them is the hypotenuse of a right angle triangle defined by the two coordinates
float MapObj::CalcH(int SrcHex,int TgtHex)
{
  HexCoordinate H1,H2;
  float L1,L2,Angle;

  H1=GetCoord(SrcHex);
  H2=GetCoord(TgtHex);

  // Check for identical Src and Tgt
  if(SrcHex==TgtHex)
    return 0.0f;

  // Calculate lengths of sides
  L1=fabs(H1.z-H2.z);
  L2=fabs(H1.x-H2.x);

  // Check for zero length sides
  if(L1==0.0f)
    return L2;

  if(L2==0.0f)
    return L1;

  // Calculate angle between hexes
  Angle=tan(L1/L2);

  // Calculate disatnce between hexes
  return (L1/sin(Angle));
}

// Find adjacent hexes of HexNum
int MapObj::GetNeigbour(int HexNum, int Which)
{
  HexPosition HexPos;

  switch(Which)
  {
  case 0: // Twelve O'Clock
    if(HexNum%RowPitch)
      return HexNum-1;
    else
      return -1; // Return -1 for top row hexes
    break;


  case 1: // Two O'Clock
    HexPos=GetPos(HexNum);
    if(HexPos.Col==HexCount/RowPitch-1) // Check for edge
      return -1;

    if(HexPos.Col%2) // Odd row
      { if(HexPos.Row==0) // Check for end of column
        return -1;
      else
        return HexNum+RowPitch-1;
    }
    else // Even row
    {
      return HexNum+RowPitch;
    }
    break;


  case 2: // Four O'Clock
    HexPos=GetPos(HexNum);
    if(HexPos.Col==HexCount/RowPitch-1) // Check for last column
      return -1;

    if(HexPos.Col%2) // Odd row
    {
      return HexNum+RowPitch;
    }
    else // Even row
    {
      if(HexPos.Row==RowPitch-1)
        return -1;
      else
        return HexNum+RowPitch+1;
    }
    break;


  case 3: // Six O'Clock
    if((HexNum-(RowPitch-1))%RowPitch)
      return HexNum+1;
    else
      return -1; // Return -1 for bottom row hexes
    break;


  case 4: // Eight O'Clock
    HexPos=GetPos(HexNum);
    if(HexPos.Col==0) // Check for edge
      return -1;

    if(HexPos.Col%2) // Odd row
    {
      return HexNum-RowPitch;
    }
    else // Even row
    {
      if(HexPos.Row==(RowPitch-1)) // Check for Last hex on column
        return -1;
      else
        return HexNum-RowPitch+1;
    }
    break;

  case 5: // Ten O'Clock
    HexPos=GetPos(HexNum);
    if(HexPos.Col==0) // Check for first column
      return -1;

    if(HexPos.Col%2) // Odd row
    {
      if(HexPos.Row==0) // First hex
        return -1;
      else
        return HexNum-RowPitch-1;
    }
    else // Even row
    {
      return HexNum-RowPitch;
    }
    break;

  default:
    return -1;
  }
}

// The actual pathfinder
int MapObj::FindPath()
{
  bool Seeking=true;
  int Loop,SrcHex,TgtHex;
  float LowF;

  // Ensure we have a start and end
  if(Start==-1 || End==-1)
    return PATH_ERR_START_END;

  if(Start==End)
    return PATH_ERR_START_END;

  // Clear out old data
  Path.clear();

  for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
  {
    HexItr->F=-1;
    HexItr->G=0.0f;
    HexItr->H=0.0f;
    HexItr->Previous=-1;
    HexItr->List=LIST_NONE;
  }


  // Add start hex to open list
  Hex[Start].List=LIST_OPEN;
  Hex[Start].F=0.f;
  Hex[Start].G=0.f;


  while(Seeking)
  {
    // Find lowest F value hex on the open list
    for(HexItr=Hex.begin(),SrcHex=-1,LowF=9999.0f,Loop=0;HexItr!=Hex.end();++HexItr,++Loop)
    {
      if(HexItr->List==LIST_OPEN)
      {
        if(HexItr->F<LowF && HexItr->F>-1)
        {
          LowF=HexItr->F;
          SrcHex=Loop;
        }
      }
    }

    // Check that we found something
    if(SrcHex==-1)
    {
      Seeking=false;
      return PATH_ERR_NO_PATH;
    }

    // Find neigbours of source hex. Set parent and F,G and H values
    for(Loop=0;Loop!=6;++Loop)
    {
      TgtHex=GetNeigbour(SrcHex,Loop);

      if(TgtHex!=-1)
      {

        // If we've found the end point, we're done searching
        if(Hex[TgtHex].Type==TYPE_FINISH)
        {
          Hex[TgtHex].List=LIST_OPEN;        // Add to open list
          Hex[TgtHex].Previous=SrcHex;       // Set parent
          Seeking=false;
          break;
        }

        // If the hex is free we'll add it to the open list
        if(Hex[TgtHex].List==LIST_NONE && Hex[TgtHex].Type!=TYPE_BLOCKED)
        {
          Hex[TgtHex].List=LIST_OPEN;        // Add to open list
          Hex[TgtHex].G=Hex[SrcHex].G+1.5f;  // Set G value (cost to here)
          Hex[TgtHex].H=CalcH(TgtHex,End);   // Set H value (cost to endpoint)
          Hex[TgtHex].F=Hex[TgtHex].G+Hex[TgtHex].H; // Set F value (G+H)
          Hex[TgtHex].Previous=SrcHex;       // Set parent
        }
        else if(Hex[TgtHex].List==LIST_OPEN) // If it's already open, check for a better route
        {
          if((Hex[SrcHex].G+10.f)<Hex[TgtHex].G) // If better G score, recalculate
          {
            Hex[TgtHex].G=Hex[SrcHex].G+1.5f;  // Set G value (cost to here)
            Hex[TgtHex].Previous=SrcHex;       // Set parent
          }
        }
      }
    }

    // Move SrcHex to closed list
    Hex[SrcHex].List=LIST_CLOSED;
  }

  // Build path
  SrcHex=End;
  while(Hex[SrcHex].Type!=TYPE_START)
  {
    Path.push_back(SrcHex); // This vector will contain the final path
    Hex[SrcHex].List=LIST_PATH;
    SrcHex=Hex[SrcHex].Previous; // Follow the previous pointer until start is reached
  }

  return PATH_OK;
}


int MapObj::FindPathStep(bool Init)
{
  static int State=STEP_START;
  static int PathHex;
  bool Seeking=true;
  int Loop,SrcHex,TgtHex;
  float LowF;

  if(Init)
  {
    Path.clear();
    State=STEP_START;
    PathHex=End;
  }

  if(Start==-1 || End==-1)
    return PATH_ERR_START_END;

  if(Start==End)
    return PATH_ERR_START_END;

  if(State==STEP_START)
  {
    // Clear out old data
    for(HexItr=Hex.begin();HexItr!=Hex.end();++HexItr)
    {
      HexItr->F=-1;
      HexItr->G=0.0f;
      HexItr->H=0.0f;
      HexItr->Previous=-1;
      HexItr->List=LIST_NONE;
    }

    // Add start hex to open list
    Hex[Start].List=LIST_OPEN;
    Hex[Start].F=0.f;
    Hex[Start].G=0.f;

    State=STEP_SEEK;
    return STEP_CONTINUE;
  }

  if(State==STEP_SEEK)
  {
    // Find lowest F value hex on the open list
    for(HexItr=Hex.begin(),SrcHex=-1,LowF=9999.0f,Loop=0;HexItr!=Hex.end();++HexItr,++Loop)
    {
      if(HexItr->List==LIST_OPEN)
      {
        if(HexItr->F<LowF && HexItr->F>-1)
        {
          LowF=HexItr->F;
          SrcHex=Loop;
        }
      }
    }

    // Check that we found something
    if(SrcHex==-1)
    {
      State=STEP_START;
      return PATH_ERR_NO_PATH;
    }

    // Find neigbours of source hex. Set parent and F,G and H values
    for(Loop=0;Loop!=6;++Loop)
    {
      TgtHex=GetNeigbour(SrcHex,Loop);

      if(TgtHex!=-1)
      {

        // If we've found the end point, we're done searching
        if(Hex[TgtHex].Type==TYPE_FINISH)
        {
          Hex[TgtHex].List=LIST_OPEN;        // Add to open list
          Hex[TgtHex].Previous=SrcHex;       // Set parent
          State=STEP_PATH;
          return STEP_CONTINUE;
        }

        // If the hex is free we'll add it to the open list
        if(Hex[TgtHex].List==LIST_NONE && Hex[TgtHex].Type!=TYPE_BLOCKED)
        {
          Hex[TgtHex].List=LIST_OPEN;        // Add to open list
          Hex[TgtHex].G=Hex[SrcHex].G+1.5f;  // Set G value (cost to here)
          Hex[TgtHex].H=CalcH(TgtHex,End);   // Set H value (cost to endpoint)
          Hex[TgtHex].F=Hex[TgtHex].G+Hex[TgtHex].H; // Set F value (G+H)
          Hex[TgtHex].Previous=SrcHex;       // Set parent
        }
        else if(Hex[TgtHex].List==LIST_OPEN) // If it's already open, check for a better route
        {
          if((Hex[SrcHex].G+10.f)<Hex[TgtHex].G) // If better G score, recalculate
          {
            Hex[TgtHex].G=Hex[SrcHex].G+1.5f;  // Set G value (cost to here)
            Hex[TgtHex].Previous=SrcHex;       // Set parent
          }
        }
      }
    }

    // Move SrcHex to closed list
    Hex[SrcHex].List=LIST_CLOSED;

    return STEP_CONTINUE;
  }

  // Build path
  Path.push_back(PathHex);  // This vector will contain the final path

  Hex[PathHex].List=LIST_PATH;
  PathHex=Hex[PathHex].Previous;  // Follow the previous pointer until start is reached

  if(Hex[PathHex].Type==TYPE_START)
    return PATH_OK;
  else
    return STEP_CONTINUE;
}
